/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package sdlip.helpers;

import sdlip.DAV;
import sdlip.DublinCore;
import sdlip.Search;
import sdlip.ResultAccess;
import sdlip.SDLIP;
import sdlip.SDLIPException;
import sdlip.http.*;
import sdlip.xml.dom.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.omg.CORBA.IntHolder;

import java.net.*;
import java.io.*;
import java.util.*;

public class ClientDienstTransport extends ClientTransportModule {

  URL lsp;
  static final String SEARCH_BOOL = "/Dienst/Index/1.0/SearchBoolean?";
  static final String DISPLAY = "/Dienst/UI/1.0/Display/";

  public ClientDienstTransport() {
  }

  public ClientDienstTransport(String lsp) throws SDLIPException {

    setLspUri(lsp);
  }

  public void setLspUri(String lspStr) throws SDLIPException {

    try {
      // name without prefix and colon
      lsp = new URL(lspStr.substring(SDLIP.DienstServerURISchema.length() + 1));
      if("http".equals(lsp.getProtocol()))
	return; // OK
    } catch (MalformedURLException mue) {}
    throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Unsupported protocol mapping: " + lsp);
  }

  /**
   * Implementation of the search method of the Abstract Interface
   */
  public void search(
		     int clientSID,
		     sdlip.XMLObject subcols,
		     sdlip.XMLObject query,
		     int numDocs,
		     sdlip.XMLObject docProps,
		     int stateTimeoutReq,
		     sdlip.XMLObject queryOptions,
		     IntHolder expectedTotal,
		     IntHolder stateTimeout,
		     IntHolder serverSID,
		     sdlip.XMLObject serverDelegate,
		     sdlip.XMLObject result
		     )
    throws SDLIPException {

    try {
      /* get Elements from anonymous XMLObjects */
      //      Element eSubcols = XMLObject.create(subcols).getElement();
      Element eQuery = XMLObject.create(query).getElement();
      Element eDocProps = XMLObject.create(docProps).getElement();
      //      Element eQueryOptions = XMLObject.create(queryOptions).getElement();

      String str = DOMUtil.getDescendantText(eQuery, DAV.contains);
      if(str == null)
	throw new SDLIPException(SDLIPException.BAD_QUERY_EXC, "Only DAV:contains operator is supported");

      str = URLEncoder.encode(str);

      str = SEARCH_BOOL + "author=" + str + "&title=" + str + "&abstract=" + str +"&boolean=or";
      /*
      if(eSubcols != null) {
	NodeList colList = eSubcols.getElementsByTagName(SDLIP.subcolName);
	for(int i=0; i < colList.getLength(); i++) {
	  String colName = DOMUtil.getText((Element)colList.item(i));
	  str += "&authority=" + colName;
	}
      }
      */
      URL req = new URL(lsp, str);
      BufferedReader in = new BufferedReader(new InputStreamReader(req.openStream()));

      /* DOM document for creation of elements */
      Document doc = XMLObject.getDocument();

      String version = in.readLine();
      if(version == null || !version.startsWith("Version"))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, "Invalid Dienst server, Version expected");

      int docCount = Integer.parseInt(in.readLine().substring(6).trim());
      if(numDocs != Search.RETURN_ALL_DOCS)
	docCount = Math.min(numDocs, docCount);
      int counter = 0;
      
      Element eResult = doc.createElement(SDLIP.SearchResult);
      // main loop
      while(counter < docCount) {
	// skip emptylines
	str = "";
	while(str != null && str.length() == 0)
	  str = in.readLine();
	// handle
	String handle = str;
	String title = in.readLine();
	String authors = in.readLine();
	String date = in.readLine();
	// construct
	Element eDoc = doc.createElement(SDLIP.doc);
	Element eDID = DOMUtil.createTextNode(doc, SDLIP.DID, String.valueOf(++counter));
	eDoc.appendChild(eDID);
	Element ePropList = doc.createElement(SDLIP.propList);
	eDoc.appendChild(ePropList);
	ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Identifier, handle) );
	try {
	  String urlStr = new URL(lsp, DISPLAY + handle.replace(':', '/')).toString();
	  ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Identifier, urlStr) );
	} catch (Exception urlex) {}
	ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Date, date) );
	ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Title, title) );
	for(StringTokenizer st = new StringTokenizer(authors, ":"); st.hasMoreTokens();) {
	  ePropList.appendChild( DOMUtil.createTextNode(doc, DublinCore.Creator, st.nextToken().trim()) );
	}
	eResult.appendChild(eDoc);
      }
      
      expectedTotal.value = counter;

      XMLObject.setElement(result, eResult);

    } catch (Exception any) {
      
      dbg.println(any);
      
      if(!(any instanceof SDLIPException))
	throw new SDLIPException(SDLIPException.SERVER_ERROR_EXC, any.getMessage());
      else
	throw (SDLIPException)any;
    }
  }

  public void cancelRequest(
			    int serverSID,
			    int reqID
			    )
    throws SDLIPException {

    throw new SDLIPException(SDLIPException.NOT_IMPLEMENTED_EXC, "");
  }


  public void getSessionInfo(
			     int serverSID,
			     IntHolder expectedTotal,
			     IntHolder stateTimeout
			     )
    throws sdlip.SDLIPException {

    throw new SDLIPException(SDLIPException.NOT_IMPLEMENTED_EXC, "");
  }
  
  public void getDocs(
		      int serverSID,
		      int reqID,
		      sdlip.XMLObject docPropList,
		      String docsToGet,
		      sdlip.XMLObject res )
    throws sdlip.SDLIPException {

    throw new SDLIPException(SDLIPException.NOT_IMPLEMENTED_EXC, "");
  }
  
  public void extendStateTimeout(
				 int serverSID,
				 int additionalTime,
				 IntHolder timeAllotted )
    throws sdlip.SDLIPException {

    throw new SDLIPException(SDLIPException.NOT_IMPLEMENTED_EXC, "");
  }

  // Metadata interface

  public void getInterface(sdlip.XMLObject intf) throws SDLIPException {

    intf.setString("<SDLIPInterface xmlns='" + SDLIP.Namespace + "'>" +
		   "  <SearchInterface/>" +
		   "  <MetadataInterface/>" +
		   "</SDLIPInterface>");
  }

  public void getSubcollectionInfo(sdlip.XMLObject subcolInfo) throws SDLIPException {

    subcolInfo.setString("<subcolInfo xmlns='" + SDLIP.Namespace + "'>" +
			 "  <subcol>" +
			 "    <subcolName>" + lsp + "</subcolName>" +
			 "    <subcolDesc>This is a collection served via the Dienst protocol</subcolDesc>" +
			 "    <defaultSubcol/>" +
			 "    <queryLangs/>" +
			 "  </subcol>" +
			 "</subcolInfo>");
  }

  public void getPropertyInfo(String subcolName, 
			      sdlip.XMLObject propInfo) throws SDLIPException {

      propInfo.setString("<propList xmlns='" + SDLIP.Namespace + "'" + 
			 "          xmlns:dc='" + DublinCore.Namespace + "'>" +
			 "  <dc:Title><searchable/><retrievable/></dc:Title>" +
			 "  <dc:Creator><searchable/><retrievable/></dc:Creator>" +
			 "  <dc:Description><searchable/><retrievable/></dc:Description>" +
			 "  <dc:Identifier><retrievable/></dc:Identifier>" +
			 "</propList>");
  }

  public String toString() {
    return "[SDLIP/Dienst Client]";
  }

  static {
    ClientTransportModule.register("dienst", "sdlip.helpers.ClientDienstTransport");
  }
}
