package sdlip.xml.dom;

import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DOMException;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public abstract class NodeImpl 
    implements Node, NodeList, Cloneable {

    //
    // Constants
    //

    //
    // Data
    //

    // properties

    /** Node name. */
	protected String name;

    /** Node value. */
	protected String value;

    // node information

    /** Owner document. */
	protected Document ownerDocument;

    /** Parent node. */
    protected NodeImpl parentNode;

    /** Previous sibling. */
	protected NodeImpl previousSibling;

    /** Next sibling. */
    protected NodeImpl nextSibling;

    /** First child. */
	protected NodeImpl firstChild;

    /** Last child. */
    protected NodeImpl lastChild;

    //
    // Constructors
    //

    protected NodeImpl(Document ownerDocument, String name, String value) {

        // set information
        this.ownerDocument = 
            (ownerDocument != null) ? ownerDocument : (Document)this;

        this.name  = name;
        this.value = value;

    } // <init>(DocumentImpl,String,short,boolean,String)

    //
    // Node methods
    //

    public String getNodeName() {

        return name;

    } // getNodeName():String

    public void setNodeValue(String value) {

    	this.value = value;

    } // setNodeValue(String)

    public String getNodeValue() {

    	return value;

    } // getNodeValue():String

    public Node appendChild(Node newChild) throws DOMException {
    	return insertBefore(newChild, null);
    }

    public Node cloneNode(boolean deep) {

    	
    	NodeImpl newnode;
    	try {
    		newnode = (NodeImpl)clone();
    	}
    	catch (CloneNotSupportedException e) {
//	        Revisit : don't fail silently - but don't want to tie to parser guts
//    		System.out.println("UNEXPECTED "+e);
    		return null;
    	}
    	
    	// Need to break the association w/ original kids
    	newnode.parentNode      = null;
    	newnode.previousSibling = null;
        newnode.nextSibling     = null;
    	newnode.firstChild      = null;
        newnode.lastChild       = null;

    	// Then, if deep, clone the kids too.
    	if (deep) {
    		for (Node child = (Node)getFirstChild(); 
    			child != null; 
    			child = (Node)child.getNextSibling()) {
    			newnode.appendChild(child.cloneNode(true));
            }
        }

        // REVISIT: Should the user data be cloned?

    	return newnode;

    } // cloneNode(boolean):Node

    public Document getOwnerDocument() {
        return ownerDocument;
    }  

    public Node getParentNode() {
        return parentNode;
    }

    /** The next child of this node's parent, or null if none */
    public Node getNextSibling() {
        return nextSibling;
    }  

    /** The previous child of this node's parent, or null if none */
    public Node getPreviousSibling() {
        return previousSibling;
    }  

    public NamedNodeMap getAttributes() {
    	return null; // overridden in ElementImpl
    }

    public boolean hasChildNodes() {
        return firstChild != null;
    }  

    public NodeList getChildNodes() {

        return this;

    } // getChildNodes():NodeList

    /** The first child of this Node, or null if none. */
    public Node getFirstChild() {

    	return firstChild;

    }   // getFirstChild():Node

    /** The first child of this Node, or null if none. */
    public Node getLastChild() {

   	    return lastChild;

    } // getLastChild():Node

    public Node insertBefore(Node newChild, Node refChild) 
        throws DOMException {

    	if( !(
    		  newChild.getOwnerDocument() == ownerDocument
    		  ||
    	 	 // SPECIAL-CASE: Document has no owner, but may be the owner.
    		  ( getNodeType() == Node.DOCUMENT_NODE &&
    		    newChild.getOwnerDocument() == (Document)this ) 
    		) ) {
    		throw new DOMExceptionImpl(DOMException.WRONG_DOCUMENT_ERR, null);
        }
        
    	// Prevent cycles in the tree
    	boolean treeSafe = true;
    	for (NodeImpl a = parentNode; treeSafe && a != null; a = a.parentNode) {
    		treeSafe = newChild != a;
        }
    	if(!treeSafe) {
    	  	throw new DOMExceptionImpl(DOMException.HIERARCHY_REQUEST_ERR, null);
        }

    	// refChild must in fact be a child of this node (or null)
    	if(refChild != null && refChild.getParentNode() != this) {
    		throw new DOMExceptionImpl(DOMException.NOT_FOUND_ERR,null);
        }
    	  
    	if (newChild.getNodeType() == Node.DOCUMENT_FRAGMENT_NODE) {
    		// SLOW BUT SAFE: We could insert the whole subtree without
    		// juggling so many next/previous pointers. (Wipe out the
    		// parent's child-list, patch the parent pointers, set the
    		// ends of the list.) But we know some subclasses have special-
    		// case behavior they add to insertBefore(), so we don't risk it.
    		// This approch also takes fewer bytecodes.

    		while (newChild.hasChildNodes())	{
    		    insertBefore(newChild.getFirstChild(), refChild);
            }
    	}
    	  
    	else {
    		Node oldparent = newChild.getParentNode();
    		if (oldparent != null) {
    			oldparent.removeChild(newChild);
            }

    		NodeImpl prev;
    		// Find the node we're inserting after, if any (null if
    		// inserting to head of list)
    		prev = (refChild == null) 
                   ? lastChild : ((NodeImpl)refChild).previousSibling;

		NodeImpl newInternal = (NodeImpl)newChild;

    		// Attach up
    		newInternal.parentNode = this;

    		// Attach after
    		newInternal.previousSibling = prev;
    		if (prev == null) {
    		    firstChild = newInternal;
            }
            else {
    		    prev.nextSibling = newInternal;
            }

    		// Attach before
    		newInternal.nextSibling = (NodeImpl)refChild;
    		if (refChild == null) {
    		    lastChild=newInternal;
            }
            else {
    		    ((NodeImpl)refChild).previousSibling = newInternal;
            }
    	}

    	return newChild;

    } // insertBefore(Node,Node):Node

    public Node removeChild(Node oldChild) 
        throws DOMException {

	if(oldChild == null)
	  return null;

    	if (oldChild.getParentNode() != this) {
    		throw new DOMExceptionImpl(DOMException.NOT_FOUND_ERR, null);
        }

    	// Patch tree past oldChild
    	NodeImpl oldInternal = (NodeImpl) oldChild;
    	NodeImpl prev = oldInternal.previousSibling;
    	NodeImpl next = oldInternal.nextSibling;

    	if (prev != null) {
    		prev.nextSibling = next;
        }
        else {
    		firstChild = next;
        }

    	if (next != null) {
    		next.previousSibling = prev;
        }
        else {
    		lastChild = prev;
        }

    	// Remove oldInternal's references to tree
    	oldInternal.parentNode      = null;
        oldInternal.nextSibling     = null;
        oldInternal.previousSibling = null;

    	return oldInternal;

    } // removeChild(Node):Node

    public Node replaceChild(Node newChild, Node oldChild) 
        throws DOMException {
    	insertBefore(newChild, oldChild);
        return removeChild(oldChild);
    }

    //
    // NodeList methods
    //

    public int getLength() {

        // It is assumed that the getChildNodes call synchronized
        // the children. Therefore, we can access the first child
        // reference directly.
    	int count = 0;
    	for (NodeImpl node = firstChild; node != null; node = node.nextSibling) {
    		count++;
    	}
    	return count;

    } // getLength():int
    
    public Node item(int index) {

        // It is assumed that the getChildNodes call synchronized
        // the children. Therefore, we can access the first child
        // reference directly.
        NodeImpl node = firstChild;
        for (int i = 0; i < index && node != null; i++) {
            node = node.nextSibling;
        }
        return node;

    } // item(int):Node

    //
    // Public methods
    //

    //
    // Object methods
    //

    public String toString() {
        return "["+getNodeName()+": "+getNodeValue()+"]";
    }    

    public abstract short getNodeType();

} // class NodeImpl
